body{
    background: #ffffff;
    color: #424247;
    font-family: sans-serif;
    font-size: 14pt;
    margin: 0;
    min-height: 100vh;
    display: flex;
    flex-wrap: wrap;
    flex-direction: row;
    align-content: center;
    align-items: flex-start;
    justify-content: space-around;
}

main{
	width: 100%;
}

.container{
    height: auto;
    min-height: 450px;
    width: 350px;
    transition: .2s;
    overflow: hidden;
    padding: 20px 40px;
    background: #fff;
    border: 1px solid #dadce0;
    border-radius: 8px;
    display: block;
    flex-shrink: 0;
    margin: 0 auto;
}

.container h1{
    margin: 0;
    width: 100%;
    text-align: center;
    color: #484848;
}

#loginscene input{
}


#loginscene .logocontainer{
    width: 100%;
    text-align: center;
}

#loginscene .logocontainer img{
    width: 75px;
}

#loginscene h1{
    text-align: center;
    font-family: sans-serif;
    font-weight: normal;
}

#loginscene button{
    float: right;
}

#loadingscene{
    width: 100%;
    height: 100%;
    background: rgb(237 237 237);
    position: absolute;
    top: 0;
    left: 0;
    display: flex;
    flex-wrap: nowrap;
    justify-content: center;
    align-items: center;
    flex-direction: column;
    overflow: hidden;
    z-index: 999;
}

#loadingscene h2{
    font-size: 2em;
    font-weight: bold;
}

#logoutview{
    width: 100%;
    display: block;
    background: white;
    text-align: center;
    padding: 10px 0px;
    color: #666;
    border-bottom: 2px solid #dadce0;
    position: fixed;
}

#logoutview span{
    width: calc(100% - 60px);
    display: inline-block;
}

#logoutview a{
    color: white;
    text-decoration: none;
    padding: 3px 10px;
    position: relative;
    border-radius: 4px;
}

#logoutview a[data-name=logout]{
    right: 25px;
    float: right;
}

#logoutview a[data-name=refresh]{
    left: 25px;
    float: left;
}

#collectionsscene{
    display: flex;
    flex-direction: row;
    flex-wrap: wrap;
    align-content: flex-start;
    align-items: center;
    margin-top: 50px;
    width: 100%;
    height: 100vh;
}

#collectionsscene article{
    width: 275px;
    background: rgb(250, 250, 250);
    border-radius: 8px;
    box-shadow: 2px 2px 3px #0000001a;
    border: 1px solid #dadce0;
    padding: 5px 10px;
    padding-top: 0;
    margin: 10px;
    float: left;
    min-height: 375px;
    overflow: hidden;
}

#collectionsscene article .colorbar{
    width: 500%;
    height: 15px;
    margin: 0px -100%;
    background: #000000;
}

#collectionsscene article .title{
    width: 100%;
    text-align: center;
    font-size: 1.5em;
    display: block;
    padding: 10px 0;
	margin: 0;
}

#collectionsscene article small{
    font-size: 15px;
    float: left;
    font-weight: normal;
    font-style: italic;
    padding-bottom: 10px;
    width: 100%;
    text-align: center;
}

#collectionsscene article input[type=text]{
    margin-bottom: 0 !important;
}

#collectionsscene article p{
    font-size: 1em;
    max-height: 130px;
    overflow: overlay;
}

#collectionsscene article:hover ul{
    visibility: visible;
}

#collectionsscene ul{
    visibility: hidden;
    display: flex;
    justify-content: space-evenly;
    width: 60%;
    margin: 0 20%;
    padding: 0;
}

#collectionsscene li{
    list-style: none;
    display: block;
}

#collectionsscene li a{
    text-decoration: none !important;
    padding: 5px;
    float: left;
    border-radius: 5px;
    width: 25px;
    height: 25px;
    text-align: center;
}

#collectionsscene article small[data-name=contentcount]{
    font-weight: bold;
    font-style: normal;
}

#editcollectionscene p span{
    word-wrap:break-word;
    font-weight: bold;
    color: #4e9a06;
}

#deletecollectionscene p span{
    word-wrap:break-word;
    font-weight: bold;
    color: #a40000;
}

#uploadcollectionscene ul{
    margin: 10px -30px;
    max-height: 600px;
    overflow-y: scroll;
}

#uploadcollectionscene li{
    border-bottom: 1px dashed #d5d5d5;
    margin-bottom: 10px;
    padding-bottom: 10px;
}

#uploadcollectionscene div[data-name=pending]{
    width: 100%;
    text-align: center;
}

#uploadcollectionscene .successmessage{
    color: #4e9a06;
    width: 100%;
    text-align: center;
    display: block;
    margin-top: 15px;
}

.deleteconfirmationtxt{
    text-align: center;
    font-size: 1em;
    font-weight: bold;
}

.fabcontainer{
    display: flex;
    flex-direction: column-reverse;
    position: fixed;
    bottom: 5px;
    right: 0;
}

.fabcontainer a{
    width: 30px;
    height: 30px;
    text-decoration: none;
    color: white;
    border: none !important;
    border-radius: 100%;
    margin: 5px 10px;
    background: black;
    text-align: center;
    display: flex;
    align-content: center;
    justify-content: center;
    align-items: center;
    font-size: 30px;
    padding: 10px;
    box-shadow: 2px 2px 7px #000000d6;
}

.title{
    word-wrap: break-word;
    font-weight: bold;
}

.icon{
    width: 100%;
    height: 100%;
    filter: invert(1);
}

.smalltext{
    font-size: 75% !important;
}

.error{
    width: 100%;
    display: block;
    text-align: center;
    color: rgb(217,48,37);
    font-family: sans-serif;
    clear: both;
    padding-top: 15px;
}

img.loading{
    width: 150px;
    height: 150px;
}

.error::before{
    content: "!";
    height: 1em;
    color: white;
    background: rgb(217,48,37);
    font-weight: bold;
    border-radius: 100%;
    display: inline-block;
    width: 1.1em;
    margin-right: 5px;
    font-size: 1em;
    text-align: center;
}

button{
    font-size: 1em;
    padding: 7px 21px;
    color: white;
    border-radius: 4px;
    float: right;
    margin-left: 10px;
    background: black;
    cursor: pointer;
}

input, select{
    width: 100%;
    height: 3em;
    border-style: solid;
    border-color: #e6e6e6;
    border-width: 1px;
    border-radius: 7px;
    margin-bottom: 25px;
    padding-left: 15px;
    padding-right: 15px;
    outline: none !important;
}

input[type=text], input[type=password]{
    width: calc(100% - 30px);
}

input:active, input:focus, input:focus-visible{
    border-color: #2494fe !important;
    border-width: 1px !important;
}

p.red, span.red{
    color: #b50202;
}

button.red, a.red{
    background: #b50202;
    border: 1px solid #a40000;
}

button.red:hover, a.red:hover{
    background: #a40000;
}

button.red:active, a.red:active{
    background: #8f0000;
}

button.green, a.green{
    background: #4e9a06;
    border: 1px solid #377200;
}

button.green:hover, a.green:hover{
    background: #377200;
}

button.green:active, a.green:active{
    background: #285200;
}

button.blue, a.blue{
    background: #2494fe;
    border: 1px solid #055fb5;
}

button.blue:hover, a.blue:hover{
    background: #1578d6;
    cursor: pointer !important;
}

button.blue:active, a.blue:active{
    background: #055fb5;
    cursor: pointer !important;
}

@media only screen and (max-width: 600px) {
    #collectionsscene{
        flex-direction: column !important;
        flex-wrap: nowrap;
    }

    #collectionsscene article{
        height: auto;
        min-height: 375px;
    }

    .container{
        max-width: 280px !important;
    }

    #collectionsscene ul{
        visibility: visible !important;
    }

    #logoutview span{
        padding: 0 5px;
    }
}
